// Property-based tests for formatter robustness and idempotency
// These tests use proptest to generate random markdown content and verify:
// 1. Rules don't crash on arbitrary input
// 2. Fixes are idempotent (applying twice gives same result)

use proptest::prelude::*;
use rumdl_lib::config::MarkdownFlavor;
use rumdl_lib::lint_context::LintContext;
use rumdl_lib::rule::{LintWarning, Rule};
use rumdl_lib::rules::*;

/// Apply all fixes from warnings to content
fn apply_all_fixes(content: &str, warnings: &[LintWarning]) -> String {
    let mut fixes: Vec<_> = warnings.iter().filter_map(|w| w.fix.as_ref()).collect();
    fixes.sort_by(|a, b| b.range.start.cmp(&a.range.start));

    let mut result = content.to_string();
    for fix in fixes {
        // Validate range is within bounds and on character boundaries
        if fix.range.end <= result.len()
            && result.is_char_boundary(fix.range.start)
            && result.is_char_boundary(fix.range.end)
        {
            result.replace_range(fix.range.clone(), &fix.replacement);
        }
    }
    result
}

/// Strategy for generating markdown-like content
fn markdown_content_strategy() -> impl Strategy<Value = String> {
    prop::collection::vec(markdown_line_strategy(), 0..20).prop_map(|lines| lines.join("\n"))
}

/// Strategy for generating individual markdown lines
fn markdown_line_strategy() -> impl Strategy<Value = String> {
    prop_oneof![
        // Headings
        (
            1..7u8,
            any::<String>().prop_filter("valid heading text", |s| { s.len() < 100 && !s.contains('\n') })
        )
            .prop_map(|(level, text)| format!("{} {}", "#".repeat(level as usize), text)),
        // List items
        any::<String>()
            .prop_filter("valid list text", |s| s.len() < 100 && !s.contains('\n'))
            .prop_map(|text| format!("- {text}")),
        // Ordered list items
        (
            1..100u32,
            any::<String>().prop_filter("valid list text", |s| { s.len() < 100 && !s.contains('\n') })
        )
            .prop_map(|(num, text)| format!("{num}. {text}")),
        // Blockquotes
        any::<String>()
            .prop_filter("valid quote text", |s| s.len() < 100 && !s.contains('\n'))
            .prop_map(|text| format!("> {text}")),
        // Code blocks
        prop::collection::vec(
            any::<String>().prop_filter("valid code", |s| s.len() < 50 && !s.contains("```")),
            0..5
        )
        .prop_map(|lines| format!("```\n{}\n```", lines.join("\n"))),
        // Inline code
        any::<String>()
            .prop_filter("valid inline code", |s| s.len() < 50 && !s.contains('`'))
            .prop_map(|text| format!("`{text}`")),
        // Links
        (
            any::<String>().prop_filter("valid link text", |s| s.len() < 30 && !s.contains(&['[', ']'][..])),
            any::<String>().prop_filter("valid url", |s| s.len() < 50 && !s.contains(&['(', ')'][..]))
        )
            .prop_map(|(text, url)| format!("[{text}]({url})")),
        // Emphasis
        any::<String>()
            .prop_filter("valid emphasis text", |s| s.len() < 50 && !s.contains('*'))
            .prop_map(|text| format!("*{text}*")),
        // Strong
        any::<String>()
            .prop_filter("valid strong text", |s| s.len() < 50 && !s.contains("**"))
            .prop_map(|text| format!("**{text}**")),
        // Plain text
        any::<String>().prop_filter("valid text", |s| s.len() < 200 && !s.contains('\n')),
        // Blank line
        Just("".to_string()),
        // Horizontal rule
        prop_oneof![
            Just("---".to_string()),
            Just("***".to_string()),
            Just("___".to_string()),
        ],
    ]
}

/// Strategy for generating completely random strings (for crash testing)
fn random_content_strategy() -> impl Strategy<Value = String> {
    any::<String>().prop_filter("reasonable size", |s| s.len() < 10000)
}

// ============================================================================
// Crash Resistance Tests
// ============================================================================

proptest! {
    #![proptest_config(ProptestConfig::with_cases(100))]

    #[test]
    fn test_lint_context_no_crash(content in random_content_strategy()) {
        // LintContext creation should never crash
        let _ = LintContext::new(&content, MarkdownFlavor::Standard, None);
        let _ = LintContext::new(&content, MarkdownFlavor::MkDocs, None);
        let _ = LintContext::new(&content, MarkdownFlavor::MDX, None);
        let _ = LintContext::new(&content, MarkdownFlavor::Quarto, None);
    }

    #[test]
    fn test_rules_no_crash(content in markdown_content_strategy()) {
        let ctx = LintContext::new(&content, MarkdownFlavor::Standard, None);

        // All these rules should never crash
        let rules: Vec<Box<dyn Rule>> = vec![
            Box::new(MD001HeadingIncrement::default()),
            Box::new(MD003HeadingStyle::default()),
            Box::new(MD004UnorderedListStyle::default()),
            Box::new(MD005ListIndent::default()),
            Box::new(MD007ULIndent::default()),
            Box::new(MD009TrailingSpaces::default()),
            Box::new(MD010NoHardTabs::default()),
            Box::new(MD011NoReversedLinks),
            Box::new(MD012NoMultipleBlanks::default()),
            Box::new(MD013LineLength::default()),
            Box::new(MD014CommandsShowOutput::default()),
            Box::new(MD018NoMissingSpaceAtx::new()),
            Box::new(MD019NoMultipleSpaceAtx),
            Box::new(MD020NoMissingSpaceClosedAtx),
            Box::new(MD021NoMultipleSpaceClosedAtx),
            Box::new(MD022BlanksAroundHeadings::default()),
            Box::new(MD023HeadingStartLeft),
            Box::new(MD024NoDuplicateHeading::default()),
            Box::new(MD025SingleTitle::default()),
            Box::new(MD026NoTrailingPunctuation::default()),
            Box::new(MD027MultipleSpacesBlockquote),
            Box::new(MD028NoBlanksBlockquote),
            Box::new(MD029OrderedListPrefix::default()),
            Box::new(MD030ListMarkerSpace::default()),
            Box::new(MD031BlanksAroundFences::default()),
            Box::new(MD032BlanksAroundLists::default()),
            Box::new(MD033NoInlineHtml::default()),
            Box::new(MD034NoBareUrls),
            Box::new(MD035HRStyle::default()),
            Box::new(MD036NoEmphasisAsHeading::default()),
            Box::new(MD037NoSpaceInEmphasis),
            Box::new(MD038NoSpaceInCode::default()),
            Box::new(MD039NoSpaceInLinks),
            Box::new(MD040FencedCodeLanguage::default()),
            Box::new(MD041FirstLineHeading::default()),
            Box::new(MD042NoEmptyLinks::default()),
            Box::new(MD045NoAltText::default()),
            Box::new(MD047SingleTrailingNewline),
            Box::new(MD049EmphasisStyle::default()),
            Box::new(MD050StrongStyle::default()),
            Box::new(MD051LinkFragments::default()),
            Box::new(MD052ReferenceLinkImages::default()),
            Box::new(MD053LinkImageReferenceDefinitions::default()),
            Box::new(MD054LinkImageStyle::default()),
            Box::new(MD055TablePipeStyle::default()),
            Box::new(MD056TableColumnCount),
            Box::new(MD058BlanksAroundTables::default()),
            Box::new(MD059LinkText::default()),
            Box::new(MD060TableFormat::default()),
            Box::new(MD061ForbiddenTerms::default()),
            Box::new(MD062LinkDestinationWhitespace),
            Box::new(MD063HeadingCapitalization::default()),
            Box::new(MD064NoMultipleConsecutiveSpaces::default()),
            Box::new(MD065BlanksAroundHorizontalRules),
            Box::new(MD066FootnoteValidation),
            Box::new(MD067FootnoteDefinitionOrder),
            Box::new(MD068EmptyFootnoteDefinition),
            Box::new(MD069NoDuplicateListMarkers),
            Box::new(MD070NestedCodeFence),
            Box::new(MD071BlankLineAfterFrontmatter),
            Box::new(MD072FrontmatterKeySort::default()),
            Box::new(MD073TocValidation::default()),
        ];

        for rule in &rules {
            let _ = rule.check(&ctx);
        }
    }
}

// ============================================================================
// Idempotency Tests
// ============================================================================

proptest! {
    #![proptest_config(ProptestConfig::with_cases(50))]

    #[test]
    fn test_md009_idempotent(content in markdown_content_strategy()) {
        let rule = MD009TrailingSpaces::default();

        let ctx1 = LintContext::new(&content, MarkdownFlavor::Standard, None);
        let warnings1 = rule.check(&ctx1).unwrap_or_default();
        let fixed1 = apply_all_fixes(&content, &warnings1);

        let ctx2 = LintContext::new(&fixed1, MarkdownFlavor::Standard, None);
        let warnings2 = rule.check(&ctx2).unwrap_or_default();
        let fixed2 = apply_all_fixes(&fixed1, &warnings2);

        prop_assert_eq!(fixed1, fixed2, "MD009 fix not idempotent");
    }

    #[test]
    fn test_md010_idempotent(content in markdown_content_strategy()) {
        let rule = MD010NoHardTabs::default();

        let ctx1 = LintContext::new(&content, MarkdownFlavor::Standard, None);
        let warnings1 = rule.check(&ctx1).unwrap_or_default();
        let fixed1 = apply_all_fixes(&content, &warnings1);

        let ctx2 = LintContext::new(&fixed1, MarkdownFlavor::Standard, None);
        let warnings2 = rule.check(&ctx2).unwrap_or_default();
        let fixed2 = apply_all_fixes(&fixed1, &warnings2);

        prop_assert_eq!(fixed1, fixed2, "MD010 fix not idempotent");
    }

    #[test]
    fn test_md012_idempotent(content in markdown_content_strategy()) {
        let rule = MD012NoMultipleBlanks::default();

        let ctx1 = LintContext::new(&content, MarkdownFlavor::Standard, None);
        let warnings1 = rule.check(&ctx1).unwrap_or_default();
        let fixed1 = apply_all_fixes(&content, &warnings1);

        let ctx2 = LintContext::new(&fixed1, MarkdownFlavor::Standard, None);
        let warnings2 = rule.check(&ctx2).unwrap_or_default();
        let fixed2 = apply_all_fixes(&fixed1, &warnings2);

        prop_assert_eq!(fixed1, fixed2, "MD012 fix not idempotent");
    }

    #[test]
    fn test_md018_idempotent(content in markdown_content_strategy()) {
        let rule = MD018NoMissingSpaceAtx::new();

        let ctx1 = LintContext::new(&content, MarkdownFlavor::Standard, None);
        let warnings1 = rule.check(&ctx1).unwrap_or_default();
        let fixed1 = apply_all_fixes(&content, &warnings1);

        let ctx2 = LintContext::new(&fixed1, MarkdownFlavor::Standard, None);
        let warnings2 = rule.check(&ctx2).unwrap_or_default();
        let fixed2 = apply_all_fixes(&fixed1, &warnings2);

        prop_assert_eq!(fixed1, fixed2, "MD018 fix not idempotent");
    }

    #[test]
    fn test_md019_idempotent(content in markdown_content_strategy()) {
        let rule = MD019NoMultipleSpaceAtx;

        let ctx1 = LintContext::new(&content, MarkdownFlavor::Standard, None);
        let warnings1 = rule.check(&ctx1).unwrap_or_default();
        let fixed1 = apply_all_fixes(&content, &warnings1);

        let ctx2 = LintContext::new(&fixed1, MarkdownFlavor::Standard, None);
        let warnings2 = rule.check(&ctx2).unwrap_or_default();
        let fixed2 = apply_all_fixes(&fixed1, &warnings2);

        prop_assert_eq!(fixed1, fixed2, "MD019 fix not idempotent");
    }

    #[test]
    fn test_md022_idempotent(content in markdown_content_strategy()) {
        let rule = MD022BlanksAroundHeadings::default();

        let ctx1 = LintContext::new(&content, MarkdownFlavor::Standard, None);
        let warnings1 = rule.check(&ctx1).unwrap_or_default();
        let fixed1 = apply_all_fixes(&content, &warnings1);

        let ctx2 = LintContext::new(&fixed1, MarkdownFlavor::Standard, None);
        let warnings2 = rule.check(&ctx2).unwrap_or_default();
        let fixed2 = apply_all_fixes(&fixed1, &warnings2);

        prop_assert_eq!(fixed1, fixed2, "MD022 fix not idempotent");
    }

    #[test]
    fn test_md031_idempotent(content in markdown_content_strategy()) {
        let rule = MD031BlanksAroundFences::default();

        let ctx1 = LintContext::new(&content, MarkdownFlavor::Standard, None);
        let warnings1 = rule.check(&ctx1).unwrap_or_default();
        let fixed1 = apply_all_fixes(&content, &warnings1);

        let ctx2 = LintContext::new(&fixed1, MarkdownFlavor::Standard, None);
        let warnings2 = rule.check(&ctx2).unwrap_or_default();
        let fixed2 = apply_all_fixes(&fixed1, &warnings2);

        prop_assert_eq!(fixed1, fixed2, "MD031 fix not idempotent");
    }

    #[test]
    fn test_md032_idempotent(content in markdown_content_strategy()) {
        let rule = MD032BlanksAroundLists::default();

        let ctx1 = LintContext::new(&content, MarkdownFlavor::Standard, None);
        let warnings1 = rule.check(&ctx1).unwrap_or_default();
        let fixed1 = apply_all_fixes(&content, &warnings1);

        let ctx2 = LintContext::new(&fixed1, MarkdownFlavor::Standard, None);
        let warnings2 = rule.check(&ctx2).unwrap_or_default();
        let fixed2 = apply_all_fixes(&fixed1, &warnings2);

        prop_assert_eq!(fixed1, fixed2, "MD032 fix not idempotent");
    }

    #[test]
    fn test_md037_idempotent(content in markdown_content_strategy()) {
        let rule = MD037NoSpaceInEmphasis;

        let ctx1 = LintContext::new(&content, MarkdownFlavor::Standard, None);
        let warnings1 = rule.check(&ctx1).unwrap_or_default();
        let fixed1 = apply_all_fixes(&content, &warnings1);

        let ctx2 = LintContext::new(&fixed1, MarkdownFlavor::Standard, None);
        let warnings2 = rule.check(&ctx2).unwrap_or_default();
        let fixed2 = apply_all_fixes(&fixed1, &warnings2);

        prop_assert_eq!(fixed1, fixed2, "MD037 fix not idempotent");
    }

    #[test]
    fn test_md038_idempotent(content in markdown_content_strategy()) {
        let rule = MD038NoSpaceInCode::default();

        let ctx1 = LintContext::new(&content, MarkdownFlavor::Standard, None);
        let warnings1 = rule.check(&ctx1).unwrap_or_default();
        let fixed1 = apply_all_fixes(&content, &warnings1);

        let ctx2 = LintContext::new(&fixed1, MarkdownFlavor::Standard, None);
        let warnings2 = rule.check(&ctx2).unwrap_or_default();
        let fixed2 = apply_all_fixes(&fixed1, &warnings2);

        prop_assert_eq!(fixed1, fixed2, "MD038 fix not idempotent");
    }

    #[test]
    fn test_md047_idempotent(content in markdown_content_strategy()) {
        let rule = MD047SingleTrailingNewline;

        let ctx1 = LintContext::new(&content, MarkdownFlavor::Standard, None);
        let warnings1 = rule.check(&ctx1).unwrap_or_default();
        let fixed1 = apply_all_fixes(&content, &warnings1);

        let ctx2 = LintContext::new(&fixed1, MarkdownFlavor::Standard, None);
        let warnings2 = rule.check(&ctx2).unwrap_or_default();
        let fixed2 = apply_all_fixes(&fixed1, &warnings2);

        prop_assert_eq!(fixed1, fixed2, "MD047 fix not idempotent");
    }

    #[test]
    fn test_md049_idempotent(content in markdown_content_strategy()) {
        let rule = MD049EmphasisStyle::default();

        let ctx1 = LintContext::new(&content, MarkdownFlavor::Standard, None);
        let warnings1 = rule.check(&ctx1).unwrap_or_default();
        let fixed1 = apply_all_fixes(&content, &warnings1);

        let ctx2 = LintContext::new(&fixed1, MarkdownFlavor::Standard, None);
        let warnings2 = rule.check(&ctx2).unwrap_or_default();
        let fixed2 = apply_all_fixes(&fixed1, &warnings2);

        prop_assert_eq!(fixed1, fixed2, "MD049 fix not idempotent");
    }

    #[test]
    fn test_md050_idempotent(content in markdown_content_strategy()) {
        let rule = MD050StrongStyle::default();

        let ctx1 = LintContext::new(&content, MarkdownFlavor::Standard, None);
        let warnings1 = rule.check(&ctx1).unwrap_or_default();
        let fixed1 = apply_all_fixes(&content, &warnings1);

        let ctx2 = LintContext::new(&fixed1, MarkdownFlavor::Standard, None);
        let warnings2 = rule.check(&ctx2).unwrap_or_default();
        let fixed2 = apply_all_fixes(&fixed1, &warnings2);

        prop_assert_eq!(fixed1, fixed2, "MD050 fix not idempotent");
    }

    #[test]
    fn test_md060_idempotent(content in markdown_content_strategy()) {
        let rule = MD060TableFormat::default();

        let ctx1 = LintContext::new(&content, MarkdownFlavor::Standard, None);
        let warnings1 = rule.check(&ctx1).unwrap_or_default();
        let fixed1 = apply_all_fixes(&content, &warnings1);

        let ctx2 = LintContext::new(&fixed1, MarkdownFlavor::Standard, None);
        let warnings2 = rule.check(&ctx2).unwrap_or_default();
        let fixed2 = apply_all_fixes(&fixed1, &warnings2);

        prop_assert_eq!(fixed1, fixed2, "MD060 fix not idempotent");
    }

    #[test]
    fn test_md062_idempotent(content in markdown_content_strategy()) {
        let rule = MD062LinkDestinationWhitespace;

        let ctx1 = LintContext::new(&content, MarkdownFlavor::Standard, None);
        let warnings1 = rule.check(&ctx1).unwrap_or_default();
        let fixed1 = apply_all_fixes(&content, &warnings1);

        let ctx2 = LintContext::new(&fixed1, MarkdownFlavor::Standard, None);
        let warnings2 = rule.check(&ctx2).unwrap_or_default();
        let fixed2 = apply_all_fixes(&fixed1, &warnings2);

        prop_assert_eq!(fixed1, fixed2, "MD062 fix not idempotent");
    }

    #[test]
    fn test_md063_idempotent(content in markdown_content_strategy()) {
        let rule = MD063HeadingCapitalization::default();

        let ctx1 = LintContext::new(&content, MarkdownFlavor::Standard, None);
        let warnings1 = rule.check(&ctx1).unwrap_or_default();
        let fixed1 = apply_all_fixes(&content, &warnings1);

        let ctx2 = LintContext::new(&fixed1, MarkdownFlavor::Standard, None);
        let warnings2 = rule.check(&ctx2).unwrap_or_default();
        let fixed2 = apply_all_fixes(&fixed1, &warnings2);

        prop_assert_eq!(fixed1, fixed2, "MD063 fix not idempotent");
    }

    #[test]
    fn test_md064_idempotent(content in markdown_content_strategy()) {
        let rule = MD064NoMultipleConsecutiveSpaces::default();

        let ctx1 = LintContext::new(&content, MarkdownFlavor::Standard, None);
        let warnings1 = rule.check(&ctx1).unwrap_or_default();
        let fixed1 = apply_all_fixes(&content, &warnings1);

        let ctx2 = LintContext::new(&fixed1, MarkdownFlavor::Standard, None);
        let warnings2 = rule.check(&ctx2).unwrap_or_default();
        let fixed2 = apply_all_fixes(&fixed1, &warnings2);

        prop_assert_eq!(fixed1, fixed2, "MD064 fix not idempotent");
    }

    #[test]
    fn test_md065_idempotent(content in markdown_content_strategy()) {
        let rule = MD065BlanksAroundHorizontalRules;

        let ctx1 = LintContext::new(&content, MarkdownFlavor::Standard, None);
        let warnings1 = rule.check(&ctx1).unwrap_or_default();
        let fixed1 = apply_all_fixes(&content, &warnings1);

        let ctx2 = LintContext::new(&fixed1, MarkdownFlavor::Standard, None);
        let warnings2 = rule.check(&ctx2).unwrap_or_default();
        let fixed2 = apply_all_fixes(&fixed1, &warnings2);

        prop_assert_eq!(fixed1, fixed2, "MD065 fix not idempotent");
    }

    #[test]
    fn test_md069_idempotent(content in markdown_content_strategy()) {
        let rule = MD069NoDuplicateListMarkers;

        let ctx1 = LintContext::new(&content, MarkdownFlavor::Standard, None);
        let warnings1 = rule.check(&ctx1).unwrap_or_default();
        let fixed1 = apply_all_fixes(&content, &warnings1);

        let ctx2 = LintContext::new(&fixed1, MarkdownFlavor::Standard, None);
        let warnings2 = rule.check(&ctx2).unwrap_or_default();
        let fixed2 = apply_all_fixes(&fixed1, &warnings2);

        prop_assert_eq!(fixed1, fixed2, "MD069 fix not idempotent");
    }

    #[test]
    fn test_md071_idempotent(content in markdown_content_strategy()) {
        let rule = MD071BlankLineAfterFrontmatter;

        let ctx1 = LintContext::new(&content, MarkdownFlavor::Standard, None);
        let warnings1 = rule.check(&ctx1).unwrap_or_default();
        let fixed1 = apply_all_fixes(&content, &warnings1);

        let ctx2 = LintContext::new(&fixed1, MarkdownFlavor::Standard, None);
        let warnings2 = rule.check(&ctx2).unwrap_or_default();
        let fixed2 = apply_all_fixes(&fixed1, &warnings2);

        prop_assert_eq!(fixed1, fixed2, "MD071 fix not idempotent");
    }
}
